Package: vasicekfit
Title: Extended Vasicek Credit Loss Model with Macroeconomic Factors
Version: 0.1.0
Authors@R: person("Dmitriy", "Mayorov", email = "cran@dimview.org",
                  role = c("aut", "cre"))
Description: Fits the extended Vasicek single-factor credit loss model where
    the probability of default depends on macroeconomic covariates. Maximum
    likelihood estimates of all parameters, including asset value correlation,
    are obtained via closed-form probit-transformed OLS regression; see
    Mayorov (2026) <doi:10.2139/ssrn.6506378> for derivation.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
Depends: R (>= 3.5.0)
Imports: stats
Suggests: testthat (>= 3.0.0), vasicek
Config/testthat/edition: 3
URL: https://github.com/externalmemory/vasicekfit-cran-package
BugReports: https://github.com/externalmemory/vasicekfit-cran-package/issues
NeedsCompilation: no
Packaged: 2026-04-23 22:22:06 UTC; default
Author: Dmitriy Mayorov [aut, cre]
Maintainer: Dmitriy Mayorov <cran@dimview.org>
Repository: CRAN
Date/Publication: 2026-04-28 18:50:02 UTC
Built: R 4.5.2; ; 2026-04-28 20:08:54 UTC; unix

export(vasicekfit)
export(dvasicek)
export(pvasicek)
export(qvasicek)
export(rvasicek)

S3method(print, vasicekfit)
S3method(summary, vasicekfit)
S3method(coef, vasicekfit)
S3method(fitted, vasicekfit)
S3method(residuals, vasicekfit)
S3method(predict, vasicekfit)
S3method(vcov, vasicekfit)
S3method(confint, vasicekfit)

importFrom(stats, coef, confint, delete.response, dnorm, lm, model.frame,
           model.matrix, model.response, pnorm, printCoefmat, qnorm, rnorm,
           terms, var, vcov)

# vasicekfit

An R package implementing the extended Vasicek single-factor credit loss
model, where the probability of default depends on macroeconomic covariates.
All parameters, including the asset value correlation, are estimated in
closed form via probit-transformed OLS.

## Installation

From CRAN (once published):

```r
install.packages("vasicekfit")
```

Development version from GitHub:

```r
# install.packages("remotes")
remotes::install_github("externalmemory/vasicekfit-cran-package")
```

## Example

```r
library(vasicekfit)

set.seed(1)
n <- 500

unemp <- rnorm(n)
hpi   <- rnorm(n)
z     <- rnorm(n)

y <- pnorm((qnorm(0.03) + 0.13 * unemp - 0.07 * hpi + sqrt(0.05) * z) / sqrt(0.95))
d <- data.frame(default_rate = y, unemp = unemp, hpi = hpi)

fit <- vasicekfit(default_rate ~ unemp + hpi, data = d)

summary(fit)
coef(fit)
confint(fit)
```

## References

Vasicek, O. A. (2002), The distribution of loan portfolio value. Risk, 15(12), 160-162.

Yang, Bill Huajian (2014), Estimating Long-Run PD, Asset Correlation, and Portfolio Level PD by Vasicek Models, MPRA Paper No. 57244 https://mpra.ub.uni-muenchen.de/57244/1/MPRA_paper_57244.pdf

Mayorov, Dmitriy, Macroeconomic Sensitivity in the Vasicek Credit Loss Model: Closed-Form Maximum Likelihood Estimation via OLS (April 01, 2026). Available at SSRN: https://ssrn.com/abstract=6506378 or http://dx.doi.org/10.2139/ssrn.6506378 

## License

MIT

#' Predict Method for vasicekfit Objects
#'
#' Obtain predictions from a fitted Vasicek model, optionally at specified
#' confidence levels.
#'
#' @param object a \code{\link{vasicekfit}} object.
#' @param newdata an optional data frame of new predictor values. If omitted,
#'   the training data are used.
#' @param type character; \code{"link"} (default) returns predictions in probit
#'   space, \code{"response"} back-transforms to the (0, 1) loss-rate scale.
#' @param alpha optional numeric vector of confidence levels in (0, 1). When
#'   supplied, predictions are conditional quantiles of the loss distribution
#'   at each confidence level. The result is a matrix with one column per
#'   alpha value.
#' @param ... additional arguments (currently unused).
#'
#' @return If \code{alpha} is \code{NULL}, a numeric vector of predictions. If
#'   \code{alpha} is supplied, a matrix with rows corresponding to observations
#'   and columns to confidence levels.
#'
#' @examples
#' set.seed(42)
#' n <- 100
#' u <- rnorm(n)
#' x <- stats::pnorm((stats::qnorm(0.03) + 0.1 * u +
#'   sqrt(0.02) * rnorm(n)) / sqrt(1 - 0.02))
#' d <- data.frame(default_rate = x, macro = u)
#' fit <- vasicekfit(default_rate ~ macro, data = d)
#'
#' # Conditional mean PD
#' predict(fit, type = "response")
#'
#' # 99th percentile loss rate under stress
#' predict(fit, newdata = data.frame(macro = 2), alpha = 0.99,
#'         type = "response")
#'
#' @export
predict.vasicekfit <- function(object, newdata = NULL,
                               type = c("link", "response"),
                               alpha = NULL, ...) {

  type <- match.arg(type)

  if (is.null(newdata)) {
    eta <- object$fitted.values
  } else {
    tt <- stats::delete.response(object$terms)
    mf <- stats::model.frame(tt, newdata)
    X <- stats::model.matrix(tt, mf)
    beta <- stats::coef(object$lm_fit)
    eta <- drop(X %*% beta)
  }


  if (is.null(alpha)) {
    pred <- switch(type,
      link     = eta,
      response = stats::pnorm(eta)
    )
    return(pred)
  }

  if (any(alpha <= 0 | alpha >= 1)) {
    stop("alpha must be in (0, 1)")
  }

  p   <- object$p
  rho <- object$rho
  kappa <- object$kappa

  if (is.null(newdata)) {
    mf_pred <- object$model
  } else {
    tt <- stats::delete.response(object$terms)
    mf_pred <- stats::model.frame(tt, newdata)
  }
  X_pred <- stats::model.matrix(stats::delete.response(object$terms), mf_pred)
  u_contrib <- drop(X_pred[, -1, drop = FALSE] %*% kappa)

  q_alpha <- stats::qnorm(alpha)
  sq_rho <- sqrt(rho)
  sq_1mrho <- sqrt(1 - rho)

  link_mat <- outer(
    stats::qnorm(p) + u_contrib,
    sq_rho * q_alpha,
    "+"
  ) / sq_1mrho

  colnames(link_mat) <- paste0("alpha_", alpha)

  pred <- switch(type,
    link     = link_mat,
    response = stats::pnorm(link_mat)
  )

  if (ncol(pred) == 1L) {
    pred <- drop(pred)
  }
  pred
}

#' The Vasicek Loss Distribution
#'
#' Density, distribution function, quantile function, and random generation
#' for the Vasicek credit loss distribution.
#'
#' @param x,q vector of quantiles (loss rates in (0, 1)).
#' @param prob vector of probabilities.
#' @param n number of observations to generate.
#' @param p probability of default, in (0, 1).
#' @param rho asset value correlation, in (0, 1).
#' @param kappa optional numeric vector of macro-factor sensitivities.
#' @param u optional numeric vector of macro-factor values, same length as
#'   \code{kappa}.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities are
#'   given as \code{log(p)}.
#' @param lower.tail logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @details
#' The Vasicek loss distribution arises from the single-factor Gaussian copula
#' model. When \code{kappa} and \code{u} are supplied, the effective PD is
#' \eqn{\Phi(\Phi^{-1}(p) + \sum \kappa_j u_j)}.
#'
#' @return \code{dvasicek} gives the density, \code{pvasicek} gives the
#'   distribution function, \code{qvasicek} gives the quantile function, and
#'   \code{rvasicek} generates random deviates.
#'
#' @examples
#' # Standard Vasicek density
#' curve(dvasicek(x, p = 0.03, rho = 0.10), from = 0.001, to = 0.20,
#'       ylab = "Density", main = "Vasicek loss distribution")
#'
#' # 99th percentile
#' qvasicek(0.99, p = 0.03, rho = 0.10)
#'
#' # Random sample
#' hist(rvasicek(10000, p = 0.03, rho = 0.10), breaks = 100)
#'
#' @name Vasicek
NULL

.vasicek_mu <- function(p, kappa = NULL, u = NULL) {
  mu <- stats::qnorm(p)
  if (!is.null(kappa) && !is.null(u)) {
    if (length(kappa) != length(u)) {
      stop("kappa and u must have the same length")
    }
    mu <- mu + sum(kappa * u)
  }
  mu
}

#' @rdname Vasicek
#' @export
dvasicek <- function(x, p, rho, kappa = NULL, u = NULL, log = FALSE) {
  if (p <= 0 || p >= 1) stop("p must be in (0, 1)")
  if (rho <= 0 || rho >= 1) stop("rho must be in (0, 1)")

  mu <- .vasicek_mu(p, kappa, u)
  z <- stats::qnorm(x)
  a <- sqrt((1 - rho) / rho)
  arg <- (sqrt(1 - rho) * z - mu) / sqrt(rho)

  log_d <- 0.5 * log((1 - rho) / rho) +
    0.5 * (z^2 - arg^2)

  # Handle boundary values
  log_d[x <= 0 | x >= 1] <- -Inf

  if (log) log_d else exp(log_d)
}

#' @rdname Vasicek
#' @export
pvasicek <- function(q, p, rho, kappa = NULL, u = NULL,
                     lower.tail = TRUE, log.p = FALSE) {
  if (p <= 0 || p >= 1) stop("p must be in (0, 1)")
  if (rho <= 0 || rho >= 1) stop("rho must be in (0, 1)")

  mu <- .vasicek_mu(p, kappa, u)

  prob <- stats::pnorm(
    (sqrt(1 - rho) * stats::qnorm(q) - mu) / sqrt(rho)
  )

  prob[q <= 0] <- 0
  prob[q >= 1] <- 1

  if (!lower.tail) prob <- 1 - prob
  if (log.p) prob <- log(prob)
  prob
}

#' @rdname Vasicek
#' @export
qvasicek <- function(prob, p, rho, kappa = NULL, u = NULL,
                     lower.tail = TRUE, log.p = FALSE) {
  if (p <= 0 || p >= 1) stop("p must be in (0, 1)")
  if (rho <= 0 || rho >= 1) stop("rho must be in (0, 1)")

  if (log.p) prob <- exp(prob)
  if (!lower.tail) prob <- 1 - prob

  mu <- .vasicek_mu(p, kappa, u)

  stats::pnorm(
    (mu + sqrt(rho) * stats::qnorm(prob)) / sqrt(1 - rho)
  )
}

#' @rdname Vasicek
#' @export
rvasicek <- function(n, p, rho, kappa = NULL, u = NULL) {
  if (p <= 0 || p >= 1) stop("p must be in (0, 1)")
  if (rho <= 0 || rho >= 1) stop("rho must be in (0, 1)")

  mu <- .vasicek_mu(p, kappa, u)
  z <- stats::rnorm(n)

  stats::pnorm((mu + sqrt(rho) * z) / sqrt(1 - rho))
}

test_that("dvasicek integrates to 1", {
  f <- function(x) dvasicek(x, p = 0.03, rho = 0.10)
  result <- integrate(f, lower = 1e-10, upper = 1 - 1e-10)
  expect_equal(result$value, 1, tolerance = 1e-6)
})

test_that("pvasicek and qvasicek are inverses", {
  p <- 0.03
  rho <- 0.10
  x_vals <- c(0.01, 0.03, 0.05, 0.10, 0.20)

  for (x in x_vals) {
    prob <- pvasicek(x, p = p, rho = rho)
    x_back <- qvasicek(prob, p = p, rho = rho)
    expect_equal(x_back, x, tolerance = 1e-10)
  }
})

test_that("pvasicek CDF matches numerical integration of dvasicek", {
  p <- 0.05
  rho <- 0.15
  q_val <- 0.08

  numerical <- integrate(
    function(x) dvasicek(x, p = p, rho = rho),
    lower = 1e-10, upper = q_val
  )$value

  analytical <- pvasicek(q_val, p = p, rho = rho)
  expect_equal(analytical, numerical, tolerance = 1e-6)
})

test_that("rvasicek sample moments match theory", {
  set.seed(42)
  p <- 0.04
  rho <- 0.08
  n <- 100000
  x <- rvasicek(n, p = p, rho = rho)

  # Theoretical mean: p
  expect_equal(mean(x), p, tolerance = 0.002)

  # All values in (0,1)
  expect_true(all(x > 0 & x < 1))
})

test_that("dvasicek with log = TRUE matches log of density", {
  x <- 0.05
  d1 <- dvasicek(x, p = 0.03, rho = 0.10)
  d2 <- dvasicek(x, p = 0.03, rho = 0.10, log = TRUE)
  expect_equal(d2, log(d1))
})

test_that("pvasicek lower.tail and log.p flags work", {
  q <- 0.05
  p <- 0.03
  rho <- 0.10

  p1 <- pvasicek(q, p, rho, lower.tail = TRUE)
  p2 <- pvasicek(q, p, rho, lower.tail = FALSE)
  expect_equal(p1 + p2, 1)

  p3 <- pvasicek(q, p, rho, log.p = TRUE)
  expect_equal(p3, log(p1))
})

test_that("qvasicek lower.tail and log.p flags work", {
  p <- 0.03
  rho <- 0.10
  prob <- 0.95

  q1 <- qvasicek(prob, p, rho, lower.tail = TRUE)
  q2 <- qvasicek(1 - prob, p, rho, lower.tail = FALSE)
  expect_equal(q1, q2)

  q3 <- qvasicek(log(prob), p, rho, log.p = TRUE)
  expect_equal(q1, q3)
})

test_that("extended distribution with kappa and u works", {
  p <- 0.03
  rho <- 0.10
  kappa <- c(0.1, -0.05)
  u <- c(1.5, 0.5)

  d <- dvasicek(0.05, p, rho, kappa = kappa, u = u)
  expect_true(is.finite(d) && d > 0)

  cdf <- pvasicek(0.05, p, rho, kappa = kappa, u = u)
  expect_true(cdf > 0 && cdf < 1)

  q99 <- qvasicek(0.99, p, rho, kappa = kappa, u = u)
  expect_true(q99 > 0 && q99 < 1)

  set.seed(1)
  r <- rvasicek(100, p, rho, kappa = kappa, u = u)
  expect_length(r, 100)
  expect_true(all(r > 0 & r < 1))
})

test_that("boundary checks for p and rho", {
  expect_error(dvasicek(0.5, p = 0, rho = 0.1), "p must be in")
  expect_error(dvasicek(0.5, p = 1, rho = 0.1), "p must be in")
  expect_error(dvasicek(0.5, p = 0.5, rho = 0), "rho must be in")
  expect_error(dvasicek(0.5, p = 0.5, rho = 1), "rho must be in")
})

test_that("dvasicek returns 0 at boundaries", {
  expect_equal(dvasicek(0, p = 0.03, rho = 0.10), 0)
  expect_equal(dvasicek(1, p = 0.03, rho = 0.10), 0)
})

test_that("kappa and u length mismatch raises error", {
  expect_error(dvasicek(0.05, 0.03, 0.1, kappa = c(0.1, 0.2), u = c(1)),
               "same length")
})

test_that("predict link matches fitted values on training data", {
  set.seed(10)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  expect_equal(predict(fit, type = "link"), fitted(fit))
})

test_that("predict response is pnorm of link", {
  set.seed(11)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  link <- predict(fit, type = "link")
  resp <- predict(fit, type = "response")
  expect_equal(resp, pnorm(link))
})

test_that("predict with newdata works", {
  set.seed(12)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  nd <- data.frame(u = c(-1, 0, 1))
  pred <- predict(fit, newdata = nd, type = "response")
  expect_length(pred, 3)
  expect_true(all(pred > 0 & pred < 1))

  # Higher u (positive kappa) should give higher PD
  expect_true(pred[3] > pred[1])
})

test_that("predict with scalar alpha returns vector", {
  set.seed(13)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  nd <- data.frame(u = c(0, 1))
  pred <- predict(fit, newdata = nd, alpha = 0.99, type = "response")
  expect_length(pred, 2)
  expect_true(all(pred > 0 & pred < 1))
})

test_that("predict with vector alpha returns matrix", {
  set.seed(14)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  nd <- data.frame(u = c(0, 1, 2))
  pred <- predict(fit, newdata = nd, alpha = c(0.95, 0.99), type = "response")
  expect_true(is.matrix(pred))
  expect_equal(dim(pred), c(3, 2))

  # Higher alpha should give higher loss rate
  expect_true(all(pred[, 2] > pred[, 1]))
})

test_that("predict alpha matches qvasicek for intercept-only model", {
  set.seed(15)
  n <- 5000
  p_true <- 0.04
  rho_true <- 0.08
  z <- rnorm(n)
  x <- pnorm((qnorm(p_true) + sqrt(rho_true) * z) / sqrt(1 - rho_true))
  d <- data.frame(y = x)
  fit <- vasicekfit(y ~ 1, data = d)

  # predict at alpha=0.99 with no covariates should approximately
  # match qvasicek with fitted p and rho
  pred99 <- predict(fit, newdata = data.frame(row.names = 1),
                    alpha = 0.99, type = "response")
  q99 <- qvasicek(0.99, p = fit$p, rho = fit$rho)
  expect_equal(pred99, q99, tolerance = 1e-10)
})

#' @export
print.vasicekfit <- function(x, ...) {
  cat("\nCall:\n")
  print(x$call)
  cat("\nVasicek model parameters:\n")
  cat("  p (PD)  =", format(x$p, digits = 6), "\n")
  cat("  rho (AVC) =", format(x$rho, digits = 6), "\n")
  if (length(x$kappa) > 0) {
    cat("  kappa:\n")
    print(x$kappa, digits = 6)
  }
  invisible(x)
}

#' @export
summary.vasicekfit <- function(object, ...) {
  cat("\nCall:\n")
  print(object$call)

  V <- stats::vcov(object)
  cf <- stats::coef(object)
  ses <- sqrt(diag(V))
  zval <- cf / ses
  pval <- 2 * stats::pnorm(-abs(zval))

  # H0: param = 0 is unreachable for p and rho (both strictly positive by
  # construction), so the Wald z / p-value are not interpretable there.
  zval[c("p", "rho")] <- NA_real_
  pval[c("p", "rho")] <- NA_real_

  ctab <- cbind(
    Estimate    = cf,
    `Std. Error` = ses,
    `z value`   = zval,
    `Pr(>|z|)`  = pval
  )

  cat("\nCoefficients (original-space parameters):\n")
  stats::printCoefmat(ctab, has.Pvalue = TRUE, na.print = "")

  cat("\nsigma2 =", format(object$sigma2, digits = 6), "\n")
  cat("Bias correction:", object$bias_correct, "\n")
  if (!is.null(object$portfolio_size)) {
    cat("Portfolio size:", object$portfolio_size, "\n")
  }

  cat("\n--- Underlying OLS regression (probit space) ---\n")
  print(summary(object$lm_fit))

  invisible(object)
}

#' @export
vcov.vasicekfit <- function(object, ...) {
  beta_hat <- stats::coef(object$lm_fit)
  sigma2   <- object$sigma2
  s        <- sqrt(1 + sigma2)
  N        <- length(object$residuals)
  k        <- length(beta_hat)
  m        <- k - 1L

  Vbeta <- stats::vcov(object$lm_fit)

  if (isTRUE(object$bias_correct)) {
    var_sigma2 <- 2 * sigma2^2 / (N - m - 1L)
  } else {
    var_sigma2 <- 2 * sigma2^2 * (N - m - 1L) / N^2
  }

  V <- matrix(0, k + 1L, k + 1L)
  V[seq_len(k), seq_len(k)] <- Vbeta
  V[k + 1L, k + 1L] <- var_sigma2

  beta0  <- beta_hat[1L]
  phi_b0 <- stats::dnorm(beta0 / s)

  J <- matrix(0, k + 1L, k + 1L)
  J[1L, 1L]      <- phi_b0 / s
  J[1L, k + 1L] <- -phi_b0 * beta0 / (2 * s^3)
  J[2L, k + 1L] <- 1 / (1 + sigma2)^2
  if (m > 0L) {
    for (j in seq_len(m)) {
      J[2L + j, 1L + j]  <- 1 / s
      J[2L + j, k + 1L] <- -beta_hat[1L + j] / (2 * s^3)
    }
  }

  out <- J %*% V %*% t(J)
  nms <- c("p", "rho", names(object$kappa))
  dimnames(out) <- list(nms, nms)
  out
}

#' @export
confint.vasicekfit <- function(object, parm, level = 0.95, ...) {
  cf  <- stats::coef(object)
  ses <- sqrt(diag(stats::vcov(object)))
  pnms <- names(cf)

  if (missing(parm)) {
    parm <- pnms
  } else if (is.numeric(parm)) {
    parm <- pnms[parm]
  }

  a <- (1 - level) / 2
  z <- stats::qnorm(1 - a)

  ci <- cbind(cf[parm] - z * ses[parm], cf[parm] + z * ses[parm])
  colnames(ci) <- sprintf("%.1f %%", 100 * c(a, 1 - a))
  rownames(ci) <- parm
  ci
}

#' @export
coef.vasicekfit <- function(object, ...) {
  c(p = object$p, rho = object$rho, object$kappa)
}

#' @export
fitted.vasicekfit <- function(object, ...) {
  object$fitted.values
}

#' @export
residuals.vasicekfit <- function(object, ...) {
  object$residuals
}

#' Fit an Extended Vasicek Credit Loss Model
#'
#' Fits the extended Vasicek single-factor credit loss model where the
#' probability of default depends on macroeconomic covariates. All parameters,
#' including asset value correlation, are estimated via closed-form
#' probit-transformed OLS.
#'
#' @param formula a formula of the form \code{response ~ predictors} where the
#'   response is an observed default/loss rate in (0, 1).
#' @param data a data frame containing the variables in \code{formula}.
#' @param bias_correct logical; if \code{TRUE}, apply the small-sample bias
#'   correction to the variance estimate by multiplying by \eqn{N/(N - m - 1)}.
#' @param portfolio_size optional positive integer. If supplied, a
#'   finite-portfolio variance correction is applied to the response before
#'   fitting (see Yang, 2014, section 4.3).
#'
#' @return An object of class \code{"vasicekfit"}, which is a list containing:
#' \describe{
#'   \item{p}{estimated probability of default (baseline PD)}
#'   \item{rho}{estimated asset value correlation}
#'   \item{kappa}{named numeric vector of macro-factor sensitivities}
#'   \item{sigma2}{MLE variance estimate in probit space}
#'   \item{lm_fit}{the underlying \code{\link{lm}} object}
#'   \item{fitted.values}{fitted values in probit (Y) space}
#'   \item{residuals}{residuals in probit (Y) space}
#'   \item{formula}{the model formula}
#'   \item{call}{the matched call}
#'   \item{terms}{the \code{\link{terms}} object}
#'   \item{model}{the model frame}
#'   \item{bias_correct}{logical flag used}
#'   \item{portfolio_size}{portfolio size used, or \code{NULL}}
#' }
#'
#' @references
#' Vasicek, O. A. (2002). The distribution of loan portfolio value.
#' \emph{Risk}, 15(12), 160--162.
#'
#' Yang, B. H. (2014). Estimating Long-Run PD, Asset Correlation, and
#' Portfolio Level PD by Vasicek Models. MPRA Paper No. 57244.
#'
#' @examples
#' set.seed(42)
#' n <- 100
#' u1 <- rnorm(n)
#' u2 <- rnorm(n)
#' p_true <- 0.03
#' rho_true <- 0.02
#' kappa_true <- c(0.13, -0.07)
#' z <- rnorm(n)
#' x <- stats::pnorm(
#'   (stats::qnorm(p_true) + kappa_true[1] * u1 + kappa_true[2] * u2 +
#'     sqrt(rho_true) * z) / sqrt(1 - rho_true)
#' )
#' d <- data.frame(default_rate = x, unemp = u1, hpi = u2)
#' fit <- vasicekfit(default_rate ~ unemp + hpi, data = d)
#' fit
#'
#' @export
vasicekfit <- function(formula, data, bias_correct = FALSE,
                       portfolio_size = NULL) {

  cl <- match.call()
  mf <- stats::model.frame(formula, data = data)
  y <- stats::model.response(mf)
  mt <- stats::terms(mf)

  if (!is.numeric(y) || any(y <= 0 | y >= 1)) {
    stop("response must be numeric with values in (0, 1)")
  }

  if (!is.null(portfolio_size)) {
    if (portfolio_size < 2) stop("portfolio_size must be >= 2")
    p0 <- mean(y)
    vr <- stats::var(y)
    v0 <- vr - (p0 * (1 - p0) - vr) / (portfolio_size - 1)
    if (v0 <= 0) stop("finite-portfolio correction yielded non-positive variance")
    y <- p0 + (y - p0) * sqrt(v0 / vr)
  }

  Y <- stats::qnorm(y)

  X <- stats::model.matrix(mt, mf)
  df_ols <- data.frame(Y = Y, X[, -1, drop = FALSE])
  lm_fit <- stats::lm(Y ~ ., data = df_ols)

  beta_hat <- stats::coef(lm_fit)
  N <- length(Y)
  sigma2 <- mean(lm_fit$residuals^2)

  if (bias_correct) {
    m <- ncol(X) - 1L
    sigma2 <- sigma2 * N / (N - m - 1L)
  }

  denom <- sqrt(1 + sigma2)
  p_hat <- stats::pnorm(beta_hat[1L] / denom)
  rho_hat <- sigma2 / (1 + sigma2)

  predictor_names <- names(beta_hat)[-1L]
  kappa_hat <- beta_hat[-1L] / denom
  names(kappa_hat) <- predictor_names

  structure(
    list(
      p              = unname(p_hat),
      rho            = rho_hat,
      kappa          = kappa_hat,
      sigma2         = sigma2,
      lm_fit         = lm_fit,
      fitted.values  = lm_fit$fitted.values,
      residuals      = lm_fit$residuals,
      formula        = formula,
      call           = cl,
      terms          = mt,
      model          = mf,
      bias_correct   = bias_correct,
      portfolio_size = portfolio_size
    ),
    class = "vasicekfit"
  )
}

test_that("vasicekfit recovers known parameters from large sample", {
  set.seed(123)
  n <- 5000
  p_true <- 0.03
  rho_true <- 0.05
  kappa_true <- c(unemp = 0.13, hpi = -0.07)

  u1 <- rnorm(n)
  u2 <- rnorm(n)
  z <- rnorm(n)
  x <- pnorm(
    (qnorm(p_true) + kappa_true[1] * u1 + kappa_true[2] * u2 +
      sqrt(rho_true) * z) / sqrt(1 - rho_true)
  )

  d <- data.frame(default_rate = x, unemp = u1, hpi = u2)
  fit <- vasicekfit(default_rate ~ unemp + hpi, data = d)

  expect_s3_class(fit, "vasicekfit")
  expect_equal(fit$p, p_true, tolerance = 0.005)
  expect_equal(fit$rho, rho_true, tolerance = 0.005)
  expect_equal(unname(fit$kappa["unemp"]), unname(kappa_true["unemp"]), tolerance = 0.01)
  expect_equal(unname(fit$kappa["hpi"]), unname(kappa_true["hpi"]), tolerance = 0.01)
})

test_that("intercept-only model recovers standard Vasicek", {
  set.seed(456)
  n <- 50000
  p_true <- 0.05
  rho_true <- 0.10
  z <- rnorm(n)
  x <- pnorm((qnorm(p_true) + sqrt(rho_true) * z) / sqrt(1 - rho_true))

  d <- data.frame(default_rate = x)
  fit <- vasicekfit(default_rate ~ 1, data = d)

  expect_equal(fit$p, p_true, tolerance = 0.02)
  expect_equal(fit$rho, rho_true, tolerance = 0.02)
  expect_length(fit$kappa, 0)
})

test_that("bias_correct adjusts sigma2 upward", {
  set.seed(789)
  n <- 50
  u <- rnorm(n)
  z <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.05) * z) / sqrt(0.95))

  d <- data.frame(y = x, u = u)
  fit_raw <- vasicekfit(y ~ u, data = d, bias_correct = FALSE)
  fit_bc <- vasicekfit(y ~ u, data = d, bias_correct = TRUE)

  expect_gt(fit_bc$sigma2, fit_raw$sigma2)
})

test_that("response must be in (0, 1)", {
  d <- data.frame(y = c(0, 0.5, 1), u = 1:3)
  expect_error(vasicekfit(y ~ u, data = d), "\\(0, 1\\)")
})

test_that("coef returns named vector with p, rho, kappa", {
  set.seed(1)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  cc <- coef(fit)
  expect_named(cc, c("p", "rho", "u"))
  expect_length(cc, 3)
})

test_that("fitted and residuals are in probit space", {
  set.seed(2)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)
  fit <- vasicekfit(y ~ u, data = d)

  expect_equal(unname(fitted(fit) + residuals(fit)), qnorm(x))
})

test_that("intercept-only fit matches vasicek::vsk_imm", {
  skip_if_not_installed("vasicek")

  set.seed(42)
  x <- vasicek::vsk_rvs(500, Rho = 0.1, P = 0.03)
  d <- data.frame(delq_rate = x)

  ref <- vasicek::vsk_imm(d$delq_rate)
  fit <- vasicekfit(delq_rate ~ 1, data = d)

  expect_equal(fit$p,   ref$P,   tolerance = 1e-12)
  expect_equal(fit$rho, ref$Rho, tolerance = 1e-12)
})

test_that("portfolio_size correction works", {
  set.seed(3)
  n <- 100
  u <- rnorm(n)
  x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(0.98))
  d <- data.frame(y = x, u = u)

  fit1 <- vasicekfit(y ~ u, data = d)
  fit2 <- vasicekfit(y ~ u, data = d, portfolio_size = 1000)

  # With finite portfolio correction, rho should be slightly different
  expect_false(identical(fit1$rho, fit2$rho))
})

simulate_fit <- function(n, p = 0.03, rho = 0.05, kappa = 0.1, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  u <- rnorm(n)
  z <- rnorm(n)
  y <- pnorm((qnorm(p) + kappa * u + sqrt(rho) * z) / sqrt(1 - rho))
  vasicekfit(y ~ u, data = data.frame(y = y, u = u))
}

test_that("vcov returns symmetric PSD matrix with correct dim and names", {
  set.seed(1)
  n <- 500
  u1 <- rnorm(n); u2 <- rnorm(n); z <- rnorm(n)
  y <- pnorm((qnorm(0.03) + 0.1 * u1 - 0.05 * u2 + sqrt(0.02) * z) / sqrt(0.98))
  fit <- vasicekfit(y ~ u1 + u2, data = data.frame(y = y, u1 = u1, u2 = u2))

  V <- vcov(fit)
  expect_equal(dim(V), c(4L, 4L))
  expect_equal(rownames(V), c("p", "rho", "u1", "u2"))
  expect_equal(colnames(V), c("p", "rho", "u1", "u2"))
  expect_equal(V, t(V))
  eig <- eigen(V, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eig >= -1e-10))
})

test_that("vcov intercept-only has 2x2 matrix for (p, rho)", {
  set.seed(10)
  n <- 500
  z <- rnorm(n)
  y <- pnorm((qnorm(0.05) + sqrt(0.1) * z) / sqrt(0.9))
  fit <- vasicekfit(y ~ 1, data = data.frame(y = y))
  V <- vcov(fit)
  expect_equal(dim(V), c(2L, 2L))
  expect_equal(rownames(V), c("p", "rho"))
})

test_that("vcov rho entry matches manual delta-method calc", {
  fit <- simulate_fit(n = 500, seed = 11)
  V <- vcov(fit)
  sigma2 <- fit$sigma2
  N <- length(fit$residuals)
  m <- length(fit$kappa)
  var_sigma2 <- 2 * sigma2^2 * (N - m - 1) / N^2
  drho_dsigma2 <- 1 / (1 + sigma2)^2
  expect_equal(V["rho", "rho"], drho_dsigma2^2 * var_sigma2)
})

test_that("confint returns matrix with correct shape, names, and contains coef", {
  fit <- simulate_fit(n = 500, seed = 2)
  ci <- confint(fit)
  expect_equal(dim(ci), c(3L, 2L))
  expect_equal(rownames(ci), c("p", "rho", "u"))
  expect_equal(colnames(ci), c("2.5 %", "97.5 %"))

  cf <- coef(fit)
  expect_true(all(ci[, 1] <= cf))
  expect_true(all(cf <= ci[, 2]))
})

test_that("confint parm subsetting works by name and index", {
  fit <- simulate_fit(n = 300, seed = 3)

  ci_rho <- confint(fit, parm = "rho")
  expect_equal(dim(ci_rho), c(1L, 2L))
  expect_equal(rownames(ci_rho), "rho")

  ci_first <- confint(fit, parm = 1)
  expect_equal(rownames(ci_first), "p")

  ci_two <- confint(fit, parm = c("p", "u"))
  expect_equal(rownames(ci_two), c("p", "u"))
})

test_that("confint level controls width", {
  fit <- simulate_fit(n = 300, seed = 4)
  w95 <- confint(fit, level = 0.95)[, 2] - confint(fit, level = 0.95)[, 1]
  w99 <- confint(fit, level = 0.99)[, 2] - confint(fit, level = 0.99)[, 1]
  expect_true(all(w99 > w95))
})

test_that("SEs scale roughly as 1/sqrt(N)", {
  n1 <- 500
  se1 <- sqrt(diag(vcov(simulate_fit(n = n1,     seed = 51))))
  se2 <- sqrt(diag(vcov(simulate_fit(n = 4 * n1, seed = 52))))
  ratio <- se1 / se2
  expect_true(all(ratio > 1.5 & ratio < 2.5))
})

test_that("Wald CI coverage is approximately nominal under simulation", {
  skip_on_cran()
  set.seed(99)
  B <- 300
  n <- 500
  p_true <- 0.03; rho_true <- 0.05; kappa_true <- 0.1

  covered <- matrix(FALSE, B, 3L)
  for (b in seq_len(B)) {
    u <- rnorm(n); z <- rnorm(n)
    y <- pnorm((qnorm(p_true) + kappa_true * u + sqrt(rho_true) * z) /
                 sqrt(1 - rho_true))
    fit <- vasicekfit(y ~ u, data = data.frame(y = y, u = u))
    ci <- confint(fit)
    covered[b, 1] <- ci["p",   1] <= p_true     && p_true     <= ci["p",   2]
    covered[b, 2] <- ci["rho", 1] <= rho_true   && rho_true   <= ci["rho", 2]
    covered[b, 3] <- ci["u",   1] <= kappa_true && kappa_true <= ci["u",   2]
  }

  cov_rates <- colMeans(covered)
  expect_true(all(cov_rates > 0.90))
  expect_true(all(cov_rates < 0.99))
})

\name{predict.vasicekfit}
\alias{predict.vasicekfit}
\title{Predict Method for vasicekfit Objects}
\description{
Obtain predictions from a fitted Vasicek model, optionally at specified
confidence levels.
}
\usage{
\method{predict}{vasicekfit}(object, newdata = NULL,
        type = c("link", "response"), alpha = NULL, \dots)
}
\arguments{
  \item{object}{a \code{\link{vasicekfit}} object.}
  \item{newdata}{an optional data frame of new predictor values. If omitted,
    the training data are used.}
  \item{type}{character; \code{"link"} (default) returns predictions in probit
    space, \code{"response"} back-transforms to the (0, 1) loss-rate scale.}
  \item{alpha}{optional numeric vector of confidence levels in (0, 1). When
    supplied, predictions are conditional quantiles of the loss distribution
    at each confidence level.}
  \item{\dots}{additional arguments (currently unused).}
}
\value{
If \code{alpha} is \code{NULL}, a numeric vector of predictions.
If \code{alpha} is supplied, a matrix with rows corresponding to observations
and columns to confidence levels. When \code{alpha} is a scalar, the result
is a vector.
}
\examples{
set.seed(42)
n <- 100
u <- rnorm(n)
x <- pnorm((qnorm(0.03) + 0.1 * u + sqrt(0.02) * rnorm(n)) / sqrt(1 - 0.02))
d <- data.frame(default_rate = x, macro = u)
fit <- vasicekfit(default_rate ~ macro, data = d)

# Conditional mean PD
predict(fit, type = "response")

# 99th percentile loss rate under stress
predict(fit, newdata = data.frame(macro = 2), alpha = 0.99,
        type = "response")

# Multiple confidence levels
predict(fit, newdata = data.frame(macro = c(0, 1, 2)),
        alpha = c(0.95, 0.99, 0.999), type = "response")
}
\seealso{
\code{\link{vasicekfit}}
}

\name{vasicekfit-package}
\alias{vasicekfit-package}
\docType{package}
\title{Extended Vasicek Credit Loss Model with Macroeconomic Factors}
\description{
Fits the extended Vasicek single-factor credit loss model where the
probability of default depends on macroeconomic covariates. Maximum likelihood
estimates of all parameters, including asset value correlation, are obtained
via closed-form probit-transformed OLS regression. Also provides density,
distribution, quantile, and random generation functions for the Vasicek loss
distribution.
}
\author{
Dmitriy Mayorov
}
\references{
Vasicek, O. A. (2002). The distribution of loan portfolio value.
\emph{Risk}, 15(12), 160--162.

Yang, B. H. (2014). Estimating Long-Run PD, Asset Correlation, and
Portfolio Level PD by Vasicek Models. MPRA Paper No. 57244.
}
\seealso{
\code{\link{vasicekfit}}, \code{\link{predict.vasicekfit}},
\code{\link{dvasicek}}
}

\name{vasicekfit-methods}
\alias{print.vasicekfit}
\alias{summary.vasicekfit}
\alias{coef.vasicekfit}
\alias{fitted.vasicekfit}
\alias{residuals.vasicekfit}
\alias{vcov.vasicekfit}
\alias{confint.vasicekfit}
\title{Methods for vasicekfit Objects}
\description{
Print, summary, coefficient extraction, covariance, confidence interval,
and accessor methods for \code{"vasicekfit"} objects.
}
\usage{
\method{print}{vasicekfit}(x, \dots)
\method{summary}{vasicekfit}(object, \dots)
\method{coef}{vasicekfit}(object, \dots)
\method{fitted}{vasicekfit}(object, \dots)
\method{residuals}{vasicekfit}(object, \dots)
\method{vcov}{vasicekfit}(object, \dots)
\method{confint}{vasicekfit}(object, parm, level = 0.95, \dots)
}
\arguments{
  \item{x, object}{a \code{\link{vasicekfit}} object.}
  \item{parm}{a specification of which parameters to compute confidence
    intervals for, either a vector of names or indices. If missing, all
    parameters are considered.}
  \item{level}{the confidence level required.}
  \item{\dots}{additional arguments (currently unused).}
}
\details{
\code{coef} returns the recovered Vasicek parameters (p, rho, and kappas)
in the original parameter space.

\code{fitted} and \code{residuals} return values in probit (Y) space,
consistent with the underlying OLS regression.

\code{vcov} returns the delta-method covariance matrix for the
original-space parameters \eqn{(p, \rho, \kappa_1, \ldots, \kappa_m)},
obtained by propagating the joint covariance of \eqn{(\hat\beta, \hat\sigma^2)}
through the closed-form recovery transform. The \eqn{(\hat\beta, \hat\sigma^2)}
block is treated as block-diagonal by the independence of sample mean and
variance under normality. The variance of \eqn{\hat\sigma^2} uses
\eqn{2 \sigma^4 (N - m - 1)/N^2} for the MLE and \eqn{2 \sigma^4 / (N - m - 1)}
when \code{bias_correct = TRUE}.

\code{confint} returns Wald intervals based on \code{vcov}.
}
\value{
\code{print} and \code{summary} return \code{object} invisibly.

\code{coef} returns a named numeric vector.

\code{fitted} and \code{residuals} return numeric vectors.

\code{vcov} returns a \eqn{(m+2) \times (m+2)} covariance matrix.

\code{confint} returns a two-column matrix of lower and upper bounds.
}
\seealso{
\code{\link{vasicekfit}}, \code{\link{predict.vasicekfit}}
}

\name{Vasicek}
\alias{dvasicek}
\alias{pvasicek}
\alias{qvasicek}
\alias{rvasicek}
\title{The Vasicek Loss Distribution}
\description{
Density, distribution function, quantile function, and random generation
for the Vasicek credit loss distribution.
}
\usage{
dvasicek(x, p, rho, kappa = NULL, u = NULL, log = FALSE)
pvasicek(q, p, rho, kappa = NULL, u = NULL, lower.tail = TRUE, log.p = FALSE)
qvasicek(prob, p, rho, kappa = NULL, u = NULL, lower.tail = TRUE, log.p = FALSE)
rvasicek(n, p, rho, kappa = NULL, u = NULL)
}
\arguments{
  \item{x, q}{vector of quantiles (loss rates in (0, 1)).}
  \item{prob}{vector of probabilities.}
  \item{n}{number of observations to generate.}
  \item{p}{probability of default, in (0, 1).}
  \item{rho}{asset value correlation, in (0, 1).}
  \item{kappa}{optional numeric vector of macro-factor sensitivities.}
  \item{u}{optional numeric vector of macro-factor values, same length as
    \code{kappa}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities are given
    as \code{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise \eqn{P[X > x]}.}
}
\details{
The Vasicek loss distribution arises from the single-factor Gaussian copula
model. Its probability density function is

\deqn{f(x) = \sqrt{\frac{1-\rho}{\rho}} \exp\left\{\frac{1}{2}\left(\Phi^{-1}(x)^2 - \left[\frac{\sqrt{1-\rho}\,\Phi^{-1}(x) - \sum \kappa_j u_j - \Phi^{-1}(p)}{\sqrt{\rho}}\right]^2\right)\right\}}{f(x) = sqrt((1-rho)/rho) * exp(0.5 * (qnorm(x)^2 - ((sqrt(1-rho)*qnorm(x) - sum(kappa * u) - qnorm(p))/sqrt(rho))^2))}

When \code{kappa} and \code{u} are \code{NULL}, this reduces to the standard
two-parameter Vasicek distribution.
}
\value{
\code{dvasicek} gives the density, \code{pvasicek} gives the distribution
function, \code{qvasicek} gives the quantile function, and \code{rvasicek}
generates random deviates.
}
\references{
Vasicek, O. A. (2002). The distribution of loan portfolio value.
\emph{Risk}, 15(12), 160--162.
}
\examples{
# Standard Vasicek density
curve(dvasicek(x, p = 0.03, rho = 0.10), from = 0.001, to = 0.20,
      ylab = "Density", main = "Vasicek loss distribution")

# 99th percentile (Basel-style)
qvasicek(0.99, p = 0.03, rho = 0.10)

# Check CDF and quantile are inverses
qvasicek(pvasicek(0.05, p = 0.03, rho = 0.10), p = 0.03, rho = 0.10)

# Random sample
set.seed(1)
hist(rvasicek(10000, p = 0.03, rho = 0.10), breaks = 100,
     main = "Vasicek loss distribution samples")
}
\seealso{
\code{\link{vasicekfit}}
}

\name{vasicekfit}
\alias{vasicekfit}
\title{Fit an Extended Vasicek Credit Loss Model}
\description{
Fits the extended Vasicek single-factor credit loss model where the
probability of default depends on macroeconomic covariates. All parameters,
including asset value correlation, are estimated via closed-form
probit-transformed OLS.
}
\usage{
vasicekfit(formula, data, bias_correct = FALSE, portfolio_size = NULL)
}
\arguments{
  \item{formula}{a formula of the form \code{response ~ predictors} where the
    response is an observed default/loss rate in (0, 1).}
  \item{data}{a data frame containing the variables in \code{formula}.}
  \item{bias_correct}{logical; if \code{TRUE}, apply the small-sample bias
    correction to the variance estimate by multiplying by \eqn{N/(N - m - 1)}, 
    where \eqn{N} is number of observations and \eqn{m} is number of predictors.}
  \item{portfolio_size}{optional positive integer. If supplied, a
    finite-portfolio variance correction is applied to the response before
    fitting (see Yang, 2014, section 4.3).}
}
\value{
An object of class \code{"vasicekfit"}, which is a list containing:
\describe{
  \item{p}{estimated probability of default (baseline PD).}
  \item{rho}{estimated asset value correlation.}
  \item{kappa}{named numeric vector of macro-factor sensitivities.}
  \item{sigma2}{MLE variance estimate in probit space.}
  \item{lm_fit}{the underlying \code{\link{lm}} object.}
  \item{fitted.values}{fitted values in probit (Y) space.}
  \item{residuals}{residuals in probit (Y) space.}
  \item{formula}{the model formula.}
  \item{call}{the matched call.}
  \item{terms}{the \code{\link{terms}} object.}
  \item{model}{the model frame.}
  \item{bias_correct}{logical flag used.}
  \item{portfolio_size}{portfolio size used, or \code{NULL}.}
}
}
\references{
Vasicek, O. A. (2002). The distribution of loan portfolio value.
\emph{Risk}, 15(12), 160--162.

Yang, B. H. (2014). Estimating Long-Run PD, Asset Correlation, and
Portfolio Level PD by Vasicek Models. MPRA Paper No. 57244.
}
\examples{
set.seed(42)
n <- 100
u1 <- rnorm(n)
u2 <- rnorm(n)
p_true <- 0.03; rho_true <- 0.02
kappa_true <- c(0.13, -0.07)
z <- rnorm(n)
x <- pnorm((qnorm(p_true) + kappa_true[1] * u1 + kappa_true[2] * u2 +
  sqrt(rho_true) * z) / sqrt(1 - rho_true))
d <- data.frame(default_rate = x, unemp = u1, hpi = u2)
fit <- vasicekfit(default_rate ~ unemp + hpi, data = d)
fit
}
\seealso{
\code{\link{predict.vasicekfit}}, \code{\link{dvasicek}}
}
